/// @file   antsontable.cpp
/// @brief  Simulation of ants walking on a table.
///
/// The table is subdivided into squares. A number of ants is
/// initially put on the table, divided as evenly as possible over the
/// squares.  Ants can move to a neighbouring square or stay put at
/// every time step.  How many of a square's ants move in each of the
/// 9 different directions is random (using the C++ random library).
/// Ants can fall of the edges of the table. The program prints the
/// number of ants left on the table after every time step, as well as
/// the minimum number of ants on any square and the maximum.
///
/// @author Ramses van Zon
/// @date   2019-2021

#include <iostream>
#include <fstream>
#include <random>

/// @brief struct to hold ant moves.
///
/// When an ant moves, their position shifts di in the i direction and
/// dj in the j direction.
///
struct direction
{
    int di; ///< shift in the 'i' direction
    int dj; ///< shift in the 'j' direction
};

const int nmoves = 9; ///< number of possible moves for an ant (3 in the i direction)x(3 in the j direction)=9 possible moves
const direction moves[nmoves] = {{-1,-1},{-1,0},{-1,1}, {0,-1}, {0,0}, {0,1}, {1,-1}, {1,0}, {1,1}}; ///< shifts in each direction for each move

/// @brief Function to compute the linear index for a matrix embedded
///        in a one-dimensional array.
///
/// @param  row   row index
/// @param  col   column index
/// @param  row   number of rows in the matrix
/// @param  ncol  number of columns in the matrix
///
/// @return the linear index
///
int linear(int row, int col, int nrow, int ncol)
{
    return row*ncol + col;
}

/// @brief Function to randomly divide a number 'total' into a set of
///        'numdivision' numbers that add up to 'total'. Uses the C++
///        random library to generate the randomness.
///
/// @param  engine    of type mt19937, to produce random numbers
/// @param  total     number that is to be subdivided (input)
/// @param  nparts    number of partitions into which to divide 'total (input)
/// @param  nperpart  resulting number in each partition (output)
///
/// @return void
///
#include <algorithm>
void rand_partition(std::mt19937& engine, int total, int nparts, int* nperpart)
{
    // Draw the points at which to 'split' the number
    std::uniform_int_distribution<int> uniformint(0, total);
    nperpart[0] = 0;
    for (int k = 1; k < nparts; k++)
        nperpart[k] = uniformint(engine);
    // By sorting, the size of each part is the just difference of
    // subsequent split points
    std::sort(&nperpart[0], &nperpart[nparts]);
    for (int k = 0; k < nparts-1; k++)
        nperpart[k] = nperpart[k+1] - nperpart[k];
    nperpart[nparts-1] = total - nperpart[nparts-1];
}

/// @brief Function to write a matrix to an already opened text file stream.
///
/// Writes each row on a separate line. Includes an trailing empty
/// line to separate subsequent calls to this function.
///
/// @param  file  an open file stream
/// @param  a     the array of integers
/// @param  nrow  number of rows in the matrix
/// @param  ncol  number of columns in the matrix
///
/// @return void
///
void write_array(std::ofstream &file, int* a, int nrow, int ncol)
{
    for (int i=0; i<nrow; i++) {
	for (int j=0; j<ncol; j++) {
	    file << a[linear(i,j,nrow,ncol)] << ' ';
	}
	file << std::endl;
    }
    file << std::endl;
}

/// @brief Main driver function of antsontable
int main()
{
    // ====================== parameters  ==================== //

    int length     = 70;     // length of the table
    int ntimesteps = 10000;  // number of time steps to take
    int total_ants = 40000;  // initial number of ants
    size_t seed    = 120;    // seed for random number generation
    std::string antsfilename = "ants.txt"; // name of file to write
    std::ofstream antsfile(antsfilename); // file stream

    // ===================== define arrays  ================== //

    // work arrays; these are linearized two-dimensional arrays
    int number_of_ants[length*length];     // distribution of ants on the table over squares.
    int new_number_of_ants[length*length]; // auxiliary array used in time step to hold the new distribution of ants
    int partition[nmoves];                 // used to determine how many ants move in which direction in a time step
    int nmin = total_ants;                 // will hold the minimum number of ants on any square
    int nmax = 0;                          // will hold the maximum number of ants on any square

    // ================= random number engine ================ //
    std::mt19937 engine(seed);

    // ==== initially place the ants evenly on the table ===== //
    for (int i = 0; i < length; i++) {
        for (int j = 0; j < length; j++) {
            int n = linear(i, j, length, length);
            number_of_ants[n] = ((long long)(n+1)*total_ants)/(length*length) - ((long long)(n)*total_ants)/(length*length);
        }
    }

    // count ants and determine minimum and maximum number on a square
    nmin = total_ants;
    nmax = 0;
    total_ants = 0;
    for (int i = 0; i < length; i++) {
        for (int j = 0; j < length; j++) {
            int n = linear(i, j, length, length); // linear index
            total_ants += number_of_ants[n];
            if (nmin > number_of_ants[n]) {
                nmin = number_of_ants[n];
            }
            if (nmax < number_of_ants[n]) {
                nmax = number_of_ants[n];
            }
        }
    }

    // ===================== start simulation ================== //

    // report and write to file
    std::cout << 0 << " " << total_ants << " " << nmin << " " << nmax << std::endl;
    write_array(antsfile, number_of_ants, length, length);

    // run time steps
    for (int timestep = 0; timestep < ntimesteps; timestep++) {

        // ants move to a new an auxiliary new 'table', empty this one first
        for (int i = 0; i < length;i++) {
            for (int j = 0; j < length;j++) {
                int n = linear(i, j, length, length);
                new_number_of_ants[n] = 0;
            }
        }

        // now move ants into auxiliary table
        for (int i = 0; i < length;i++) {
            for (int j = 0; j < length;j++) {
                int n = linear(i, j, length, length);
                if (number_of_ants[n] > 0 ) {
                    // pick how many ants take part in each of the 9 moves
                    rand_partition(engine, number_of_ants[n], nmoves, partition);
                    // push ants in their respective move directions
                    for (int m = 0; m < nmoves; m++) {
                        int i2 = i + moves[m].di;
                        int j2 = j + moves[m].dj;
                        // put only on new table if not falling off table
                        if (i2>=0 and i2<length and j2>=0 and j2<length) {
                            int n = linear(i2, j2, length, length);
                            new_number_of_ants[n] += partition[m];
                        }
                    }
                }
            }
        }

        // update number_of_ants
        for (int i = 0; i < length; i++) {
            for (int j = 0; j < length; j++) {
                int n = linear(i, j, length, length); // linear index
                number_of_ants[n] = new_number_of_ants[n];
            }
        }

        // count ants and determine minimum and maximum number on a square
        nmin = total_ants;
        nmax = 0;
        total_ants = 0;
        for (int i = 0; i < length; i++) {
            for (int j = 0; j < length; j++) {
                int n = linear(i, j, length, length);
                total_ants += number_of_ants[n];
                if (nmin > number_of_ants[n]) {
                    nmin = number_of_ants[n];
                }
                if (nmax < number_of_ants[n]) {
                    nmax = number_of_ants[n];
                }
            }
        }

        // report and write to file
        std::cout << timestep+1 << " " << total_ants << " " << nmin << " " << nmax << std::endl;
	write_array(antsfile, number_of_ants, length, length);
    }

    antsfile.close();
    return 0;
}
